use std::net::SocketAddr;
use domo_proto::packet::identifier::Identifier;
use std::io;
use tokio::net::{TcpSocket, TcpStream};
use domo_proto::commands::node_management::NodeManagement;
use std::io::ErrorKind;
use domo_proto::packet::{Packet, ToPacket};
use tokio::io::AsyncWriteExt;
use crate::connection;

pub struct Client {
    socket_addr: SocketAddr,
}

impl Client {
    pub fn new(socket_addr: SocketAddr) -> Self {
        Self {
            socket_addr
        }
    }

    pub async fn client_connection(self, device_id: Option<Identifier>) -> io::Result<ClientConnection> {
        ClientConnection::new(self, device_id).await
    }
}

pub struct ClientConnection {
    stream: TcpStream,
    device_id: Identifier,
}

impl ClientConnection {

    async fn register_node(stream: &mut TcpStream, device_id: Identifier) -> io::Result<Identifier> {
        stream.write(
            NodeManagement::RegisterNode { device_id }
                .to_packet(
                    Identifier::default(),
                    Identifier::default(),
                    Identifier::random(),
                    Identifier::default(),
                )
                .build_full_packet()
                .as_slice()
        ).await?;
        let response = connection::from_tcp_stream(stream).await?;
        if response.command == 0x01 {
            if Identifier::from(response.data.data.clone()) != device_id {
                return Err(io::Error::new(ErrorKind::InvalidData, "Server did not send appropiate response."))
            }
            Ok(Identifier::from(response.data.data))
        } else {
            Err(io::Error::new(ErrorKind::AddrNotAvailable, "Address not available"))
        }
    }

    pub async fn new(c: Client, device_id: Option<Identifier>) -> io::Result<Self> {
        let sock = TcpSocket::new_v4()?;
        let mut stream = sock.connect(c.socket_addr).await?;

        let device_id = match device_id {
            Some(device_id) => ClientConnection::register_node(&mut stream, device_id),
            None => ClientConnection::register_node(&mut stream, Identifier::default())
        }.await?;

        Ok(Self {
            stream,
            device_id
        })
    }

    pub fn stream(&self) -> &TcpStream { &self.stream }
    pub fn device_id(&self) -> Identifier { self.device_id }

    pub async fn send(&mut self, packet: Packet) -> io::Result<Packet> {
        self.stream.write(packet.build_full_packet().as_slice()).await?;
        connection::from_tcp_stream(&mut self.stream).await
    }
}
